unit capture;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics;

function CaptureWindow(hWndSrc: HWND;
  Client: Boolean;
  LeftSrc, TopSrc, WidthSrc, HeightSrc: Integer): TBitmap;

implementation

// From http://support.microsoft.com/kb/q161299

// qqq leaks one GDI handle somewhere.
function CaptureWindow(hWndSrc: HWND;
  Client: Boolean;
  LeftSrc, TopSrc, WidthSrc, HeightSrc: Integer): TBitmap;
var
  hDCMemory: Integer;
  hBmp: Integer;
  hBmpPrev: Integer;
  hDCSrc: Integer;
  hPal: Integer;
  hPalPrev: Integer;
  RasterCapsScrn: Integer;
  HasPaletteScrn: Boolean;
  PaletteSizeScrn: Integer;
  LogPal: LOGPALETTE;
  bmp: TBitmap;

begin
  Result := nil;
  hBmp := 0;
  hPal := 0;
  hDCSrc := 0;
  hDCMemory := 0;
  bmp := nil;
  hPalPrev := 0;

  try
    try
      // Depending on the value of Client get the proper device context.
      if Client then begin
        // Get device context for client area.
        hDCSrc := GetDC(hWndSrc);
      end else begin
        // Get device context for entire window
        hDCSrc := GetWindowDC(hWndSrc);
      end;
      if hDCSrc = 0 then Exit;

      try
        // Create a memory device context for the copy process.
        hDCMemory := CreateCompatibleDC(hDCSrc);
        if hDCMemory = 0 then Exit;

        // Create a bitmap and place it in the memory DC.
        hBmp := CreateCompatibleBitmap(hDCSrc, WidthSrc, HeightSrc);
        hBmpPrev := SelectObject(hDCMemory, hBmp);

        // Get screen properties. Raster capabilities...
        RasterCapsScrn := GetDeviceCaps(hDCSrc, RASTERCAPS);
        // Palette support...
        HasPaletteScrn := (RasterCapsScrn and RC_PALETTE) <> 0;
        // Size of palette.
        PaletteSizeScrn := GetDeviceCaps(hDCSrc, SIZEPALETTE);

        // If the screen has a palette make a copy and realize it.
        if HasPaletteScrn and (PaletteSizeScrn = 256) then begin
          // Create a copy of the system palette.
          LogPal.palVersion := $300;
          LogPal.palNumEntries := 256;
          GetSystemPaletteEntries(hDCSrc, 0, 256, LogPal.palPalEntry[0]);
          hPal := CreatePalette(LogPal);
          // Select the new palette into the memory DC and realize it.
          hPalPrev := SelectPalette(hDCMemory, hPal, False);
          RealizePalette(hDCMemory);
        end;

        // Copy the on-screen image into the memory DC.
        BitBlt(hDCMemory, 0, 0, WidthSrc, HeightSrc, hDCSrc, LeftSrc, TopSrc, SRCCOPY);

        // Remove the new copy of the  on-screen image.
        hBmp := SelectObject(hDCMemory, hBmpPrev);

        // If the screen has a palette get back the palette that was
        // selected in previously.
        if HasPaletteScrn and (PaletteSizeScrn = 256) then begin
          hPal := SelectPalette(hDCMemory, hPalPrev, False);
        end;

      finally
        // Release the device context resources back to the system.
        DeleteDC(hDCMemory);
      end;

    finally
      ReleaseDC(hWndSrc, hDCSrc);
    end;

    // Done - hBmp & hPal are the result. Now create a TBitmap and return..
    try
      bmp := TBitmap.Create;
      bmp.Handle := hBmp;
      bmp.Palette := hPal;
      bmp.Width := WidthSrc;
      bmp.Height := HeightSrc;

      Result := TBitmap.Create;
      if Assigned(Result) then Result.Assign(bmp);

      // Modify the bitmap to ensure it's copied.
      with Result.Canvas do begin
        MoveTo(0,0);
        LineTo(1,1);
      end;
    finally
      bmp.Free;
    end;

  finally
    if hBmp <> 0 then DeleteObject(hBmp);
    if hPal <> 0 then DeleteObject(hPal);
  end;

end;

end.

