unit main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  OleCtrls, GFFCONTROLLib_TLB, StdCtrls, ExtCtrls, Buttons;

const
  WM_FIXKEYDOWN = WM_USER + 3678;
  WM_FIXKEYUP = WM_USER + 3679;

  BaseCaption = 'Genline';

type
  TMainForm = class(TForm)
    Button1: TButton;
    SetupTimer: TTimer;
    CaptionTimer: TTimer;
    BitBtn1: TBitBtn;
    procedure Button1Click(Sender: TObject);
    procedure SetupTimerTimer(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure CaptionTimerTimer(Sender: TObject);
    procedure BitBtn1Click(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
  private
    GFFinder: TGFFinder;
    LoginHnd: HWND;
    MainHnd: HWND;
    ResultHnd: HWND;
    SettingsHnd: HWND;
    MainSettingsBtnHnd: HWND; // Main - knappen Instllningar
    LightTrackBar: HWND;
    Forward5: HWND;
    Back5: HWND;
    IsFullscreen: Boolean;
    ShiftState: TShiftState;
    MainThreadId: Cardinal;
    GffThreadId: Cardinal;
    ZoomCount: Integer;
    SetupAttempts: Integer;
    MaxAttempts: Integer;

    // For mouse capture
    start_x, start_y: Integer;
    end_x, end_y: Integer;
    mouseCanvas: TCanvas;
    capturingMouse: Boolean;
    wantToCaptureMouse: Boolean;

    Initializing: Boolean;

    function CaptureCleanup: Boolean;
    procedure WMKeyDown(var Message: TWMKeyDown); message WM_FIXKEYDOWN;
    procedure WMKeyUp(var Message: TWMKeyUp); message WM_FIXKEYUP;
    procedure WMLButtonDown(var Message: TWMMouse); message WM_LBUTTONDOWN;
    procedure WMLButtonUp(var Message: TWMMouse); message WM_LBUTTONUP;
    procedure WMMouseMove(var Message: TWMMouse); message WM_MOUSEMOVE;
    procedure HandleKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure ShowHelp;
    procedure DoCaptureWindow(x, y, w, h: Integer);
  public
    procedure CreateGFF(AOwner: TForm);
  end;

var
  MainForm: TMainForm;

implementation

{$R *.DFM}
uses
  Registry, DebugWnd, Capture, CaptWnd, ShellApi, Math;

type
  TMyRegistry = class(TRegistry)
  public
    function ReadIntegerDef(const Name: string; default: Integer): Integer;
  end;

procedure DebugOutputStr(const s: string; args: array of const); forward;

type
  findwin_context = record
    className: string;
    caption: string;
    hnd: HWND;
  end;
  pfindwin_context = ^findwin_context;

var
  winHook: HHOOK;

function hookProc(nCode: Integer; W: WPARAM; L: LPARAM): Integer; stdcall;
begin
  // DebugOutputStr('Ho', []);
  if nCode = HC_ACTION then begin
    if (L and $80000000) = 0 then
      PostMessage(MainForm.Handle, WM_FIXKEYDOWN, W, L)
    else
      PostMessage(MainForm.Handle, WM_FIXKEYUP, W, L);
  end;
  Result := CallNextHookEx(winHook, nCode, W, L);
end;

procedure DebugOutputStr(const s: string; args: array of const);
begin
  // OutputDebugString(PChar(Format(s + #10, args)));
  DebugWindow.DebugList.Lines.Add(Format(s, args));
end;

{ TMyRegistry }

function TMyRegistry.ReadIntegerDef(const Name: string; default: Integer): Integer;
begin
  Result := default;
  try
    Result := ReadInteger(Name);
  except
  end;
end;

{ TMainForm }

procedure TMainForm.CreateGFF(AOwner: TForm);
begin
  winHook := SetWindowsHookEx(WH_KEYBOARD, hookProc, 0, GetCurrentThreadId);
  if (winHook = 0) then
    DebugOutputStr('Hooking keyboard failed', []);

  GFFinder := TGFFinder.Create(AOwner);
  with GFFinder do begin
    Parent := AOwner;
    Left := 0;
    Top := 0;
    Align := alClient;
  end;

  SetupAttempts := 0;
  MaxAttempts := 15;
  Initializing := True;
  SetupTimer.Enabled := true;
  GFFinder.Show;
end;


procedure TMainForm.Button1Click(Sender: TObject);
begin
  if not Assigned(CapturedData) then begin
    CapturedData := TCapturedData.Create(nil);
  end;
  CreateGFF(Self);
end;

function FindChildWindow(h: HWND; const cls: string; const caption: string): HWND;
var
  context: findwin_context;
  function EnumFunc(hnd: HWND; L: LPARAM): Boolean; stdcall;
  var
    ourClass, ourName: string;
  type
    PHWND = ^HWND;
  begin
    Result := True;
    SetLength(ourClass, 512);
    GetClassName(hnd, PChar(ourClass), 512);
    ourClass := PChar(ourClass);
    SetLength(ourName, 512);
    GetWindowText(hnd, PChar(ourName), 512);
    ourName := PChar(ourName);
    { DebugOutputStr('Child Class="%s" Name="%s" Search="%s"/"%s"', [ourClass, ourName,
      pfindwin_context(L)^.className, pfindwin_context(L)^.caption]); }
    if (ourClass = pfindwin_context(L)^.className) and
      (ourName = pfindwin_context(L)^.caption) then
    begin
      pfindwin_context(L)^.hnd := hnd;
      // DebugOutputStr('Found window %x', [hnd]);
      Result := False;
    end;
  end;
begin
  context.className := cls;
  context.caption := caption;
  context.hnd := 0;
  EnumChildWindows(h, @EnumFunc, Longint(@context));
  // DebugOutputStr('Result %x', [context.hnd]);
  Result := context.hnd;
end;

procedure TMainForm.WMKeyDown(var Message: TWMKeyDown);
begin
  // DebugOutputStr('Down=%x', [Message.Charcode]);
  if Message.Charcode = VK_SHIFT then
    ShiftState := ShiftState + [ssShift]
  else if Message.Charcode = VK_CONTROL then
    ShiftState := ShiftState + [ssCtrl]
  else if Message.Charcode = VK_MENU then
    ShiftState := ShiftState + [ssAlt]
  else
    HandleKeyDown(Self, Message.Charcode, ShiftState);

  inherited;
end;

procedure TMainForm.WMKeyUp(var Message: TWMKeyUp);
begin
  // DebugOutputStr('Up=%x', [Message.Charcode]);
  if Message.Charcode = VK_SHIFT then
    ShiftState := ShiftState - [ssShift]
  else if Message.Charcode = VK_CONTROL then
    ShiftState := ShiftState - [ssCtrl]
  else if Message.Charcode = VK_MENU then
    ShiftState := ShiftState - [ssAlt];
  if Windows.GetKeyState(VK_SHIFT) = 0 then ShiftState := ShiftState - [ssShift];
  if Windows.GetKeyState(VK_CONTROL) = 0 then ShiftState := ShiftState - [ssCtrl];
  if Windows.GetKeyState(VK_MENU) = 0 then ShiftState := ShiftState - [ssAlt];
  inherited;
end;


procedure TMainForm.WMLButtonDown(var Message: TWMMouse);
begin
  if not wantToCaptureMouse then Exit;
  wantToCaptureMouse := False;
  capturingMouse := True;
  start_x := Message.XPos;
  start_y := Message.YPos;
  end_x := Message.XPos;
  end_y := Message.YPos;
  mouseCanvas := TCanvas.Create;
  mouseCanvas.Handle := GetDC(GFFinder.Handle);
  with mouseCanvas.Pen do begin
    Color := clBlack;
    Mode := pmXor;
    Width := 1;
  end;
  mouseCanvas.Rectangle(start_x, start_y, end_x, end_y);
end;

procedure TMainForm.WMMouseMove(var Message: TWMMouse);
begin
  if not capturingMouse then Exit;
  if not Assigned(mouseCanvas) then Exit;
  mouseCanvas.Rectangle(start_x, start_y, end_x, end_y);
  end_x := Message.XPos;
  end_y := Message.YPos;
  mouseCanvas.Rectangle(start_x, start_y, end_x, end_y);
end;

function TMainForm.CaptureCleanup: Boolean;
begin
  Result := False;
  if not capturingMouse then Exit;
  if not Assigned(mouseCanvas) then Exit;
  mouseCanvas.Rectangle(start_x, start_y, end_x, end_y);
  ReleaseCapture;
  capturingMouse := False;
  wantToCaptureMouse := False;
  mouseCanvas.Free;
  mouseCanvas := nil;
  Result := True;
end;

procedure TMainForm.WMLButtonUp(var Message: TWMMouse);
  procedure swap(var x, y: Integer);
  var
    tmp: Integer;
  begin
    tmp := y;
    y := x;
    x := tmp;
  end;
begin
  if not CaptureCleanup then Exit;
  end_x := Message.XPos;
  end_y := Message.YPos;
  if (start_x > end_x) then swap(start_x, end_x);
  if (start_y > end_y) then swap(start_y, end_y);
  DoCaptureWindow(start_x, start_y, end_x - start_x, end_y - start_y);
end;

procedure TMainForm.ShowHelp;
begin
  ShellExecute(0, 'open', 'brux.htm', nil, nil, SW_SHOW);
  (*MessageDlg('Fljande knappar kan anvndas.'#13#10 +
    'F1 - visa denna text'#13#10 +
    'F5 - starta kopiering med musen'#13#10 +
    'F6 - visa fnstret med kopierad bild'#13#10 +
    'F9 - visa bildfnstret'#13#10 +
    'F10 - visa Skresultatet'#13#10 +
    'F11 - visa GFF (dr man vljer frsamling osv.)'#13#10 +
    'PageUp - g till fregende sida'#13#10 +
    'PageDown - g till nsta sida'#13#10 +
    'Ctrl + PageUp - g fem sidor bakt'#13#10 +
    'Ctrl + PageDown - g fem sidor framt'#13#10 +
    ''#13#10 +
    'P numeriska tangentbordet:'#13#10 +
    'Minus - zooma ut (mindre)'#13#10 +
    'Plus - zooma in (strre)'#13#10 +
    'Division - mrkare'#13#10 +
    'Multiplikation - ljusare'#13#10 +
    ''#13#10 +
    'Piltangenter: flytta bilden'#13#10
    , mtInformation, [mbOK], 0);*)
end;

procedure TMainForm.DoCaptureWindow(x, y, w, h: Integer);
var
  bmp: TBitmap;
  pict: TPicture;
begin
  bmp := nil;
  pict := nil;
  try
    bmp := CaptureWindow(GFFinder.Handle, True, x, y, w, h);
    if not Assigned(bmp) then Exit;
    try
      pict := TPicture.Create;
      pict.Bitmap := bmp;
      with CapturedData.Image1 do begin
        Picture := pict;
        Width := w;
        Height := h;
      end;
    finally
      pict.Free;
    end;
  finally
    bmp.Free;
  end;
  with CapturedData do begin
    ImageHeight := h;
    ImageWidth := w;
    Show;
  end;
end;

procedure TMainForm.HandleKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  case Key of
    VK_ESCAPE:
      begin
        ShiftState := [];
        CaptureCleanup;
      end;

    VK_F1:
      begin
        ShiftState := [];
        ShowHelp;
      end;

    VK_F2:
      begin
        DebugWindow.Show;
      end;

    VK_F5:
      begin
        wantToCaptureMouse := True;
        SetCapture(Self.Handle);
      end;

    VK_F6:
      begin
        CapturedData.Show;
      end;

    VK_F9:
      begin
        MainForm.Activate;
      end;

    VK_F10:
      begin
        AttachThreadInput(GffThreadId, MainThreadId, true);
        SetActiveWindow(ResultHnd);
        Application.ProcessMessages;
        AttachThreadInput(GffThreadId, MainThreadId, false);
        //PostMessage(ResultHnd, WM_LBUTTONDOWN, 0, 0);
        //PostMessage(ResultHnd, WM_LBUTTONUP, 0, 0);
      end;

    VK_F11:
      begin
        AttachThreadInput(GffThreadId, MainThreadId, true);
        SetActiveWindow(MainHnd);
        Application.ProcessMessages;
        AttachThreadInput(GffThreadId, MainThreadId, false);
        //stat := PostMessage(MainHnd, WM_LBUTTONDOWN, MK_LBUTTON, $10001);
        //DebugOutputStr('Ho=%x', [GetLastError]);
        //stat := PostMessage(MainHnd, WM_LBUTTONUP, MK_LBUTTON, $10001);
        //DebugOutputStr('Ho=%x', [GetLastError]);
      end;

    { Fullscreen; Inactivated - GFF will recreate its window, so we have to login again.
    VK_F12: begin
        if IsFullscreen then begin
          IsFullscreen := false;
          WindowState := wsNormal;
          //Application.ProcessMessages;
          //BorderStyle := bsSizeable;
          //Application.ProcessMessages;
        end else begin
          IsFullscreen := true;
          //BorderStyle := bsNone;
          WindowState := wsMaximized;
          //Application.ProcessMessages;
          //BorderStyle := bsSingle;
          //Application.ProcessMessages;
        end;
      end;}

    VK_F12:
      begin
        if ssCtrl in Shift then begin
          // Roterande positionslista hr
        end else begin
          with MainForm do begin
            Left := 0;
            Top := 0;
            Width := Screen.Width;
            Height := Screen.Height;
          end;
        end;
      end;


    VK_PRIOR:
      begin
        if ssCtrl in Shift then begin
          Application.ProcessMessages;
          PostMessage(Back5, WM_LBUTTONDOWN, 0, 0);
          PostMessage(Back5, WM_LBUTTONUP, 0, 0);
          Sleep(100);
          Application.ProcessMessages;
          Sleep(100);
        end;
      end;

    VK_NEXT: begin
        if ssCtrl in Shift then begin
          Application.ProcessMessages;
          PostMessage(Forward5, WM_LBUTTONDOWN, 0, 0);
          PostMessage(Forward5, WM_LBUTTONUP, 0, 0);
          Sleep(100);
          Application.ProcessMessages;
          Sleep(100);
        end;
      end;

    { Lighter }
    VK_MULTIPLY:
      begin
        PostMessage(LightTrackBar, WM_KEYDOWN, VK_RIGHT, 0);
        PostMessage(LightTrackBar, WM_KEYUP, VK_RIGHT, Integer($80000000));

      end;

    { Darker }
    VK_DIVIDE:
      begin
        PostMessage(LightTrackBar, WM_KEYDOWN, VK_LEFT, 0);
        PostMessage(LightTrackBar, WM_KEYUP, VK_LEFT, Integer($80000000));
      end;

    { Zoom out - handled by GFF }
    VK_SUBTRACT:
      begin
        Dec(ZoomCount);
        if (ZoomCount < 0) then ZoomCount := 0;
      end;

    { Zoom in - handled by GFF }
    VK_ADD:
      begin
        Inc(ZoomCount);
        if (ZoomCount > 20) then ZoomCount := 20;
      end;

  end;
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  IsFullscreen := False;
end;


procedure TMainForm.SetupTimerTimer(Sender: TObject);
var
  h: HWND;
  i: Integer;
  allOK: Boolean;
  r: TMyRegistry;
  zoom: Integer;
  rect: TRect;
begin
  zoom := 0;
  allOK := true;
  SetupTimer.Enabled := false;

  { The login window }
  LoginHnd := FindWindow('TLoginForm', 'Login');
  if (LoginHnd = 0) then begin
    DebugOutputStr('TLoginForm FAILED', []);
    allOK := false;
  end;

  { Genline Family Finder window }
  MainHnd := FindWindow('TMainForm', 'Genline Family Finder');
  if (MainHnd = 0) then begin
    DebugOutputStr('TMainForm (GFF) FAILED', []);
    allOK := false;
  end;

  { Search Result window }
  ResultHnd := FindWindow('TResultForm', 'Skresultat');
  if (ResultHnd = 0) then begin
    DebugOutputStr('TResultForm (Sk) FAILED', []);
    allOK := false;
  end;

  { The Settings dialog }
  SettingsHnd := FindWindow('TSettingsForm', 'Instllningar');
  if (SettingsHnd = 0) then begin
    DebugOutputStr('TSettingsForm FAILED', []);
    allOK := false;
  end;

  { The "Instllningar" button }
  MainSettingsBtnHnd := FindChildWindow(MainHnd, 'TButton', 'Instllningar');
  if MainSettingsBtnHnd = 0 then begin
    DebugOutputStr('SettBtn FAILED', []);
    allOK := false;
  end;

  { Now "click" the Instllningar and wait for the Settings window
  to materialize. Will take a few seconds. }
  Application.ProcessMessages;
  PostMessage(MainSettingsBtnHnd, WM_LBUTTONDOWN, 0, 0);
  PostMessage(MainSettingsBtnHnd, WM_LBUTTONUP, 0, 0);
  Sleep(100);
  Application.ProcessMessages;
  Sleep(100);

  i := 0;
  h := 0;
  repeat
    Inc(i);
    if i > MaxAttempts then Break;

    { Find the Close button on the Settings window }
    h := FindChildWindow(SettingsHnd, 'TButton', '&Stng');
    if (h <> 0) then begin
      Application.ProcessMessages;
      PostMessage(h, WM_LBUTTONDOWN, 0, 0);
      PostMessage(h, WM_LBUTTONUP, 0, 0);
      Sleep(100);
      Application.ProcessMessages;
      Sleep(100);
    end;
  until h <> 0;

  if h = 0 then begin
    DebugOutputStr('Hittar inte Stng-knappen i instllningsfnstret', []);
    allOK := false;
  end;

  { The Gamma trackbar }
  LightTrackBar := FindChildWindow(SettingsHnd, 'TTrackBar', '');
  if LightTrackBar = 0 then begin
    DebugOutputStr('TTrackBar (Light) FAILED', []);
    allOK := false;
  end;

  { The >> button in the GFF window }
  Forward5 := FindChildWindow(MainHnd, 'TButton', '>>');
  if Forward5 = 0 then allOK := False;

  { The << button in the GFF window }
  Back5 := FindChildWindow(MainHnd, 'TButton', '<<');
  if Back5 = 0 then allOK := False;

  { Find the thread identifiers }
  MainThreadId := GetWindowThreadProcessId(MainForm.Handle, nil);
  if MainThreadId = 0 then allOK := False;
  GffThreadId := GetWindowThreadProcessId(MainHnd, nil);
  if GffThreadId = 0 then allOK := False;

  { Connect thread input queues }
  // qqq gra det permant har elaka sidoeffekter d smfnstrena poppar upp d & d.
  // stat := AttachThreadInput(GffThreadId, MainThreadId, true);
  // DebugOutputStr('ATI: %x', [Integer(stat)]);

  Inc(SetupAttempts);
  if not allOK then begin
    if (SetupAttempts < MaxAttempts) then begin
      SetupTimer.Enabled := True;
    end else begin
      DebugOutputStr('En del fnster vill inte visa sig.', []);
      DebugOutputStr('Har dock tjatat p dem %d gnger.', [SetupAttempts]);
      // qqq MessageBox hr?
    end
  end;

  if allOK then begin
    r := TMyRegistry.Create;
    try
      r.RootKey := HKEY_CURRENT_USER;
      if r.OpenKey('\Software\DanielBerglund\GenlineSkal', False) then begin

        Left := r.ReadIntegerDef('X', Left);
        Top := r.ReadIntegerDef('Y', Top);
        Width := r.ReadIntegerDef('W', Width);
        Height := r.ReadIntegerDef('H', Height);

        Left := Min(Left, Screen.Width - 5);
        Top := Min(Top, Screen.Height - 5);

        { Set the position of the GFF window }
        try
          GetWindowRect(MainHnd, rect);
          rect.Left := r.ReadIntegerDef('gX', rect.Left);
          rect.Top := r.ReadIntegerDef('gY', rect.Top);
          rect.Left := Min(rect.Left, Screen.Width - 5);
          rect.Top := Min(rect.Top, Screen.Height - 5);
          SetWindowPos(MainHnd, HWND_TOP, rect.Left, rect.Top, 0, 0, SWP_SHOWWINDOW + SWP_NOSIZE);
        finally
        end;

        { qqq funkar ibland - fixa}
        try zoom := r.ReadInteger('Zoom'); finally end;
        for zoom := zoom downto 1 do begin
          Sleep(20);
          PostMessage(GFFinder.Handle, WM_KEYDOWN, VK_ADD, 0);
          Sleep(20);
          PostMessage(GFFinder.Handle, WM_KEYUP, VK_ADD, Integer($80000000));
        end;

      end;
      Initializing := False;

    finally
      r.CloseKey;
      r.Free;
    end;
  end;

  GFFinder.Show;
end;


procedure TMainForm.CaptionTimerTimer(Sender: TObject);
var s: string;
begin
  if Initializing then begin
    s := 'Initialiserar - var god vnta...';
  end else if Assigned(GFFinder) then begin
    s := BaseCaption + Format(' GID %s', [string(GFFinder.gid)]);
  end else begin
    s := BaseCaption;
  end;
  if Caption <> s then Caption := s;
  // DebugOutputStr('Gid="%s" Gidinfo="%s"', [string(GFFinder.gid), string(GFFinder.gidinfo)]);
end;


procedure TMainForm.BitBtn1Click(Sender: TObject);
begin
  ShowHelp;
end;

procedure TMainForm.FormClose(Sender: TObject; var Action: TCloseAction);
var
  r: TRegistry;
  rect: TRect;
begin
  r := TRegistry.Create;
  try
    r.RootKey := HKEY_CURRENT_USER;
    if r.OpenKey('\Software\DanielBerglund\GenlineSkal', True) then begin
      r.WriteInteger('X', MainForm.Left);
      r.WriteInteger('Y', MainForm.Top);
      r.WriteInteger('W', MainForm.Width);
      r.WriteInteger('H', MainForm.Height);
      r.WriteInteger('Zoom', ZoomCount);

      GetWindowRect(MainHnd, rect);
      r.WriteInteger('gX', rect.Left);
      r.WriteInteger('gY', rect.Top);

    end;
  finally
    r.CloseKey;
    r.Free;
  end;
end;

end.

